/*
 *
 * Copyright (C) 2010 Martin Bosslet <Martin.Bosslet@googlemail.com>
 * All rights reserved.
 */
/*
 * This program is licenced under the same licence as Ruby.
 * (See the file 'COPYING'.)
 */

#if !defined(_OSSL_TS_H_)
#define _OSSL_TS_H_

void Init_ossl_ts(void);

#endif
